/* simply include tin's autoconf'd header file */
#ifndef TIN_AUTOCONF_H
#	include "autoconf.h"
#endif /* !TIN_AUTOCONF_H */
#define EBCDIC 0
#define POSIX_MALLOC_THRESHOLD 10
#define PCRE_EXPORT
#define MAX_NAME_SIZE 32
#define MAX_NAME_COUNT 10000
#define MAX_DUPLENGTH 30000
